import AutoRPE.UtilsRPE.VariablePrecision as VariablePrecision
import AutoRPE.UtilsRPE.RegexPattern as RegexPattern
import AutoRPE.UtilsRPE.Error as Error
import re


def load_sources(path):
    from AutoRPE.UtilsRPE.BasicStructures import ReadSourceFile
    from os.path import join, abspath
    from os import listdir
    ord_list = listdir(path)
    ord_list.sort()
    file_list = [ReadSourceFile(join(path, filename)) for filename in
                 ord_list if
                 filename.count(".f90") and filename[0][0] != "."]
    if not len(file_list):
        raise ValueError("No files found in %s folder", abspath(path))
    return file_list


def literal_mask(string):
    splitted_string = [c for c in string]
    length = len(splitted_string)
    mask = []
    inside_literal = False
    marker = None
    for i in range(length):
        if splitted_string[i] == "\"":
            if not inside_literal:
                inside_literal = True
                marker = "\""
                mask.append(False)
            else:
                if marker == "\"":
                    inside_literal = False
                    marker = None
                    mask.append(False)
                elif marker == "'":
                    mask.append(True)
                else:
                    raise AssertionError("Should not happen!")
        elif splitted_string[i] == "'":
            if not inside_literal:
                inside_literal = True
                marker = "'"
                mask.append(False)
            else:
                if marker == "'":
                    inside_literal = False
                    marker = None
                    mask.append(False)
                elif marker == "\"":
                    mask.append(True)
                else:
                    raise AssertionError("Should not happen!")
        else:
            if not inside_literal:
                mask.append(False)
            else:
                mask.append(True)

    return mask


def remove_literals(string):
    mask = literal_mask(string)
    modified_string = "".join(
        [c if not mask[i] else " " for i, c in enumerate(string)])
    return modified_string


def replace_structures(string):
    array = re.findall(RegexPattern.array, string, re.I)
    purged_string = string
    if array:
        array_store = []
        for index, a in enumerate(array):
            array_store.append(a)
            purged_string = purged_string.replace(a, "ARRAY" + str(index))
    structure = re.findall(RegexPattern.structure, purged_string, re.I)
    if structure:
        for s in structure:
            # Keep the length of the string constant, so that regex doesn't fail
            substitution = ""
            if s.count("ARRAY"):
                array_number = re.findall(r"\barray(\d+)", s, re.I)
                len_s = len(s)
                for number in array_number:
                    len_s -= len("ARRAY"+number)
                    substitution += "#"*len(array_store[int(number)])
                substitution += "#"*len_s
            else:
                substitution = "#"*len(s)
            # Don't replace with a word to avoid regex is_call_to_function to match
            purged_string = purged_string.replace(s, substitution)
        for index, a in enumerate(array):
            purged_string = purged_string.replace("ARRAY" + str(index), array_store[index])
        return purged_string
    return string


def close_brackets(string):
    started = False
    bracket_counter = 0
    mask = literal_mask(string)

    for _i, _c in enumerate(string):
        if _c == "(" and not mask[_i]:
            bracket_counter += 1
            if not started:
                started = True
        elif _c == ")" and not mask[_i]:
            bracket_counter -= 1
            if not started:
                raise ValueError("This should not happen, check arguments %s" % string)
        if bracket_counter < 0:
            raise ValueError("This should not happen, check arguments %s" % string)
        if started and bracket_counter == 0:
            return string[:_i + 1]

    if string.count("(") + string.count(")") == 0:
        return string
    raise ValueError("This should not happen, check arguments %s" % string)


def combine_types(list_of_types):

    ret_val = list_of_types.pop()
    while list_of_types:
        next_type = list_of_types.pop()
        try:
            ret_val = VariablePrecision.dictionary_of_possible_combination[frozenset([ret_val, next_type])]
        except KeyError:
            raise Error.ExceptionNotManaged("Please add the combination" +
                                            "[" + str(ret_val) + "," + str(next_type) + "]" +
                                            " to dictionary_of_possible_combination ")
    return ret_val


def create_type_combinations(case_types):
    """
    Given a list of variable types, it creates all the possible combinations casting real types.
    :param case_types: list of variable types
    :return: list of all the possible variations of variable types
    """
    # TODO: In some cases a big number of arguments can make this approach crash. Here we add a security check.
    if len(case_types) > 10:
        raise AssertionError("Too much arguments to use a combinatorial approach.")
    import itertools
    import itertools
    syn_types = ["real" if x in VariablePrecision.real_id else x for x in case_types]

    number_of_reals = syn_types.count("real")

    a = ["sp", "dp"] * number_of_reals
    combinations = list(set(itertools.permutations(a, number_of_reals)))

    new_types = []
    for comb in combinations:
        comb_types = syn_types[:]
        for i in range(number_of_reals):
            comb_types[comb_types.index("real")] = comb[i]
        new_types.append(comb_types)
    return new_types


def find_next_end(lines, _index, end_statement):
    import re
    """ 
    Given a bunch of lines and a line number, looks for next line with the string contained in end_statement argument.
    Input:
          _index: integer with line number
          end_statement: string that indicates next end_statement
    Output: integer with the line number of the next line containing the end_statement
    """
    _i = _index + 1
    begin_statement = end_statement.replace('end', '')
    nested_declaration = 0
    while _i < len(lines):
        _l = lines[_i]
        if re.match('^\s*(' + begin_statement + ')', _l, re.I):
            nested_declaration += 1
        if _l.lower().count(end_statement):
            if nested_declaration == 0:
                return _i
            else:
                nested_declaration -= 1
        _i += 1

    return len(lines)


def is_all_inside_parenthesis(string):
    if string.strip()[0] == "(":
        try:
            if "c%s" % string.strip() == close_brackets("c%s" % string.strip()):
                return True
            else:
                return False
        except ValueError:
            return False
    else:
        return False


def remove_comments(string):
    # Trying to use a rarely used character ({) for the temporal replacement
    if string.count("!"):
        if string.count('$OMP'):
            return string
        # In the case where there's quotes in the text, we should ignore the exclamations inside
        if not string.count("'") and not string.count('"'):
            string = string[:string.find("!")]
        else:
            mask = literal_mask(string)
            # Looking for
            splitted_string = [c for c in string]
            splitted_string = ["{" if c == "!" and mask[i] else c for i, c in enumerate(splitted_string)]
            string = "".join(splitted_string)

            if string.find("!") >= 0:
                string = string[:string.find("!")]
            string = string.replace("{", "!")

    return string


def remove_if_condition(_line, return_if_condition=False):
    if _line.replace(" ", "").lower().strip().find("if(") == 0:
        condition = close_brackets(_line)
        _line = _line.replace(condition, "")
    else:
        condition = ""
    if not return_if_condition:
        return _line
    else:
        return [condition, _line]


def split_if_condition(_line):
    return remove_if_condition(_line, return_if_condition=True)


def remove_outer_brackets(string):
    string = string.strip()
    if is_all_inside_parenthesis(string):
        return string[1:-1].strip()
    else:
        return string


def remove_empty_lines(_lines):
    """
      Remove empty strings from list of strings
    """
    return [_l for _l in _lines if _l.strip()]


# Used in split_elements_in_operations
def remove_brackets(string):
    replacement_index = 0
    arrays = True
    replacement_list = []

    if string.strip()[0] == "[" and string.strip()[-1] == "]":
        string = string.strip()[1:-1]
    # Checks for all array pattern, and substitute with string
    while arrays:
        arrays = re.findall(RegexPattern.array, string)
        for array in arrays:
            string = string.replace(array, "REPLACEMENT" + str(replacement_index), 1)
            replacement_list.append(array)
            replacement_index += 1

        # Substitute structures with string
        structures = re.findall(r"\w+%\w+", string)
        for struct in structures:
            string = string.replace(struct, "REPLACEMENT" + str(replacement_index))
            replacement_list.append(struct)
            replacement_index += 1
            arrays = True
    functions = re.search(r"\w+\s*\(", string)
    if functions:
        matches = re.finditer(r"\w+\s*\(", string, flags=re.I)
        matches_start = [m.start(0) for m in matches]
        for start in matches_start:
            func = close_brackets(string[start:])
            string = string.replace(func, "REPLACEMENT" + str(replacement_index))
            replacement_list.append(func)
            replacement_index += 1

    # Now is safe to discard brackets
    string = string.replace("(", " ")
    string = string.replace(")", " ")

    # Substitute back the original structures
    for index, replacement in reversed(list(enumerate(replacement_list))):
        string = string.replace("REPLACEMENT" + str(index), replacement)

    return string


def merge_lines(_lines):
    """
    Given a file, puts back on the same line commands split with a &
    """
    for _index, _line in enumerate(_lines):

        line = _lines[_index]
        if not line:
            continue
        if line[0] == "#":
            continue
        if line.count("&"):
            i = 1
            while line and line.strip()[-1] == "&":
                line1 = line
                line2 = _lines[_index + i]
                line = line1.replace("&", "").rstrip() + line2
                i += 1
            line = line.replace("&", "")
            _lines[_index] = line
            # # A commands to delete unnecessary spaces (gives some problems right now)
            # line = re.sub(", +", ", ", line)
            # line = re.sub(":  +", ":  ", line)
            # _lines[_index] = line
            for j in range(1, i):
                _lines[_index + j] = ""
    _lines = remove_empty_lines(_lines)
    return _lines


# Merge calls split on multiple lines and remove comments
def merge_line_range(lines, line_index):
    line = remove_comments(lines[line_index]).strip()

    merge_line_range.number = 1
    if line.count("&"):
        while line and line.strip()[-1] == "&":
            line1 = line
            while remove_comments(lines[line_index + merge_line_range.number]).strip() == "":
                merge_line_range.number = merge_line_range.number + 1
            line2 = remove_comments(lines[line_index + merge_line_range.number])
            line = line1.replace("&", "").rstrip() + line2
            merge_line_range.number += 1
        line = line.replace("&", "")
        line = re.sub(", +", ", ", line)
        line = re.sub(":  +", ":  ", line)
        line = line.strip() + "\n"

    return line


# Merge lines in file
def merge_line_in_file(file, line_index):
    with open(file, "w") as f:
        lines = [l for l in f]
        line = merge_line_range(lines, line_index)
        lines[line_index] = line
        for j in range(1, merge_line_range.number):
            lines[line_index + j] = ""

        text = "".join(lines)
        f.write(text)
