import AutoRPE.UtilsRPE.Error as Error


def generate_dict_key(el_types, el_dimensions, uniform=False):
    import AutoRPE.UtilsRPE.VariablePrecision as VariablePrecision
    # Create a key agnostic on the real type
    if uniform:
        # Perform a copy not to modify the args in input
        el_types = el_types[:]
        for index, t in enumerate(el_types):
            if t in list(VariablePrecision.real_id) + ["rpe_var"]:
                el_types[index] = "uniform_real"

    key = [len(el_types)] + \
          el_types + \
          el_dimensions
    return hash(str(key))


class ReadSourceFile:
    def __init__(self, _filename):
        # Init variables
        self.path = _filename
        self.filename = self.path.split("/")[-1]
        self.module_name = self.filename.replace(".f90", "")
        self.lines = []
        self.text = ""

        # Read file
        self.read_file(_filename)

    def read_file(self, _filename):
        try:
            with open(_filename) as f:
                self.text = f.read()
                self.lines = [l.strip() for l in self.text.split("\n")]
        except UnicodeDecodeError:
            raise UnicodeDecodeError("The file %s contains non UTF-8 character.\n"
                                     " Please clean, and retry.", self.path)

    def rebuild_text(self):
        self.text = "".join([l + "\n" for l in self.lines[:] if l.strip()])

    def save_file(self, file_path=None):
        if file_path:
            path = file_path + "/" + self.filename
        else:
            path = self.path
        with open(path, "w") as output_file:
            output_file.write(self.text)


class Variable:
    def __init__(self, procedure, _name, _type, _dimension, _module=None):
        self.id = None
        self.module = _module
        self.procedure = procedure
        self.name = _name
        self.type = _type
        self.dimension = _dimension
        self.is_dummy_argument = False
        self.intent = None
        self.is_optional = None
        self.is_retval = False
        self.is_pointer = False
        self.position = None
        self.allocatable = False
        self.appears_in_namelist = False
        self.precision = None
        self.is_parameter = None
        self.is_member_of = None
        self.mutable = None
        self.same_as = []


class Procedure:
    def __init__(self, _name, module):
        self.name = _name
        self.module = module
        self.variables = []
        self.precision = None
        self.interface = None
        self.scope = None
        self.declares = None
        self.dummy_arguments = []
        self.uses = []
        self.accessible_var = {}
        self.variables_dictionary = {}

    def add_variable(self, _variable):
        try:
            assert isinstance(_variable, Variable)
        except AssertionError:
            pass
        _variable.procedure = self
        self.variables.append(_variable)

    def get_variable_by_position(self, position):
        for variable in self.dummy_arguments:
            if variable.position == position:
                return variable
        raise Error.VariableNotFound("Variable not found in position %d" % position)

    def get_variable_by_name(self, name):
        # In the pre-processing phase this dictionary is empty
        if self.variables_dictionary:
            try:
                return self.variables_dictionary[name]
            except KeyError:
                None
        else:
            for variable in self.variables:
                if variable.name.lower() == name.lower():
                    return variable

    def fill_dictionary_of_variable(self):
        for var in self.variables:
            self.variables_dictionary[var.name] = var

    def fill_dictionary_of_accessible_var(self):

        if isinstance(self, DerivedType):
            # TODO probably for extended structures I have to add the extension variables
            for var in self.variables:
                self.accessible_var[var.name] = [var]
            return

        for var in self.variables + self.module.variables + self.module.accessible_var:
            try:
                self.accessible_var[var.name] += [var]
            except KeyError:
                self.accessible_var[var.name] = [var]
        if self.scope:
            for var in self.scope.variables:
                try:
                    self.accessible_var[var.name] += [var]
                except KeyError:
                    self.accessible_var[var.name] = [var]
        try:
            # Main, Interface and TYPE do not have this attribute
            for var in self.use_only_variables:
                try:
                    self.accessible_var[var.name] += [var]
                except KeyError:
                    self.accessible_var[var.name] = [var]
        except AttributeError:
            pass


class DerivedType(Procedure):
    def __init__(self, _name, module, _extended_structure=None, is_external=False):
        super(DerivedType, self).__init__(_name, module)
        self.is_external = is_external
        self.extended_structure = _extended_structure
        self.extension = None

    def get_variable(self, item):
        for variable in self.variables:
            if variable.name.lower() == item.lower():
                return variable


class Module(ReadSourceFile):
    def __init__(self, file_name):
        super(Module, self).__init__(file_name)
        self.name = self.module_name
        self.main = None
        self.blocks = []
        self.variables = []
        self.precision = None
        self.uses = []
        self.used_by = []
        self.accessible_var = []
        self.indirect_accessible_var = []

    def add_variable(self, _variable):
        assert isinstance(_variable, Variable)
        _variable.procedure = self
        self.variables.append(_variable)

    def update_blocks(self, vault):
        import AutoRPE.UtilsRPE.BasicFunctions as BasicFunctions
        import AutoRPE.UtilsRPE.RegexPattern as RegexPattern
        import re
        # Routine to know if we are entering or exiting a procedure
        # If the line corresponds to the start of a procedure, returns the name of this procedure
        # If the line corresponds to the end of a procedure, returns None
        # If is neither one of these cases, returns the same procedure
        block = self.main
        parent_block = None
        # Rebuild the text
        text = "".join([l + "\n" for l in self.lines[:] if l.strip()])
        self.lines = text.split("\n")
        # Reset the list
        self.blocks = []
        for _index, _line in enumerate(self.lines):
            _line = _line.strip()
            _line = BasicFunctions.remove_literals(_line)
            _pattern = None
            if _line.lower().count("subroutine"):
                _pattern = RegexPattern.subroutine_declaration
            elif _line.lower().count("function"):
                _pattern = RegexPattern.function_declaration
            elif _line.lower().find("type") == 0:
                _pattern = RegexPattern.simple_type_declaration
            # We are changing block
            if _pattern is not None:
                # We are exiting a block
                if _line.lower().find("end") == 0:
                    # Check we are not exiting a nested function
                    if parent_block is not None:
                        if parent_block == block:
                            parent_block = None
                            block = None
                        else:
                            block = parent_block
                    else:
                        block = None

                # We are entering a new block
                else:
                    match = re.search(_pattern, _line, flags=re.IGNORECASE)
                    # cases like variables names that contains "function" in it will not match
                    if match is not None:
                        block_name = match.groups()[0].strip()

                        try:
                            block = vault.get_procedure_by_name(block_name, self.module_name)
                        except Error.ProcedureNotFound:
                            block = None

                        if parent_block is None:
                            parent_block = block
            # For every line, stores the block which the line belongs to
            self.blocks.append(block)


class Interface:
    def __init__(self, _name, module):
        self.name = _name
        self.module = module
        self.subprogram = []
        self.key = {}
        self.agnostic_key = {}

    def add_key(self, module_procedure):
        mandatory_da = [dummy for dummy in module_procedure.dummy_arguments if not dummy.is_optional]
        mandatory_da_type = [da.type for da in mandatory_da]
        mandatory_da_dim = [da.dimension for da in mandatory_da]
        key = generate_dict_key(mandatory_da_type, mandatory_da_dim)
        agnostic_key = generate_dict_key(mandatory_da_type, mandatory_da_dim, uniform=True)
        # Create a unique key from interface argument starting from attribute
        try:
            self.key[key] += [module_procedure.name]
        except KeyError:
            self.key[key] = [module_procedure.name]
        try:
            self.agnostic_key[agnostic_key] = [module_procedure.name]
        except KeyError:
            self.agnostic_key[agnostic_key] += [module_procedure.name]

    def add_procedure(self, subprogram):
        assert isinstance(subprogram, SubRoutine) or isinstance(subprogram, Function)

        self.subprogram.append(subprogram)
        # Don't know it will be useful yet
        if not subprogram.interface:
            subprogram.interface = self
        elif isinstance(subprogram.interface, list):
            subprogram.interface.append(self)
        else:
            subprogram.interface = [subprogram.interface, self]


class SubRoutine(Procedure):
    def __init__(self, _name, module, is_external=False):
        super(SubRoutine, self).__init__(_name, module)
        # External subroutine will just have this attribute to true
        self.is_external = is_external
        # Case in which a use statement is found inside a subroutine definition
        self.use_only_variables = []


class Function(Procedure):
    def __init__(self, _name, module, is_external=False):
        super(Function, self).__init__(_name, module)
        self.is_external = is_external
        self.type = "external" if is_external else None
        self.use_only_variables = []


class Argument:
    def __init__(self, name, type=None, var=None):
        self.name = name
        self.type = type
        self.variable = var


class SubprogramCall:
    def __init__(self, call, subprogram):
        self.call = call
        self.subprogram = subprogram
        self.arguments = []
        self.dummy_arguments = []
