import AutoRPE.UtilsRPE.BasicStructures as BasicStructures
import AutoRPE.UtilsRPE.BasicFunctions as BasicFunctions
import AutoRPE.UtilsRPE.Splitter as Splitter
import AutoRPE.UtilsRPE.CallManager as CallManager
import AutoRPE.UtilsRPE.NatureDeterminer as NatureDeterminer
import AutoRPE.UtilsRPE.VariablePrecision as VariablePrecision
import AutoRPE.UtilsRPE.Error as Error
import AutoRPE.UtilsRPE.RegexPattern as RegexPattern

import re
import warnings


def get_variable(contents, procedure=None, vault=None):
    contents = contents.replace(" ", "").strip()

    contents = BasicFunctions.remove_literals(contents)

    if NatureDeterminer.has_intrinsics(contents):
        raise Error.VariableNotFound("The string %s is not a variable" % contents)
    if NatureDeterminer.is_an_hardcoded_array(contents):
        raise Error.VariableNotFound("The string %s is not a variable" % contents)

    if NatureDeterminer.is_logical_comparison(contents):
        raise Error.VariableNotFound("The string %s is not a variable" % contents)

    if NatureDeterminer.has_operations(contents):
        raise Error.VariableNotFound("The string %s is not a variable" % contents)

    # Remove the indexing from array and DT:
    #   ght_abl <-- ght_abl(2:jpka)
    #   todelay%z1d <-- todelay(ji)%z1d(:)
    contents = remove_indexing(contents)
    if re.search(RegexPattern.variable, contents):
        try:
            return vault.get_variable_by_name(var_name=contents, procedure=procedure)
        except Error.VariableNotFound:
            raise Error.VariableNotFound

    if contents.count("%"):
        if contents.count("%val"):
            contents = contents.replace("%val", "")
            return get_variable(contents, procedure, vault)
        # Search for all others members
        try:
            return get_structure_member(contents, procedure, vault)
        except Error.MemberNotFound:
            raise Error.VariableNotFound

    if is_keyword_argument(contents):
        # Removing before first =
        contents = contents[contents.find("=") + 1:].strip()
        return get_variable(contents, procedure, vault)

    raise Error.VariableNotFound("Variable %s: not found" % contents)


def get_real_in_line(line, vault, procedure, intent=["out", "inout"]):
    subprogram_call = CallManager.get_procedure_inline(line, vault, procedure)
    ret_dummy = []
    ret_var = []
    if not subprogram_call:
        if NatureDeterminer.is_pointer_assignment(line):
            ptr, target = get_pointer_target(line, procedure, vault)
            if [ptr, target] != [None, None] and ptr.type in VariablePrecision.real_id:
                # This is needed because we do not have an argument/dummy but we need to copy the 'else' structure
                arg_trg = BasicStructures.Argument(line.split("=>")[1], target.type, target)
                return [[ptr], [arg_trg]]
    else:
        for s_call in subprogram_call:
            if not s_call.subprogram.is_external:
                for dummy_a, arg in zip(s_call.dummy_arguments, s_call.arguments):
                    if dummy_a.intent in intent:
                        if dummy_a.type in VariablePrecision.real_id:
                            ret_dummy.append(dummy_a)
                            ret_var.append(arg)
    # Check something was found
    if len(ret_dummy):
        return [ret_dummy, ret_var]
    return [None, None]


def get_pointer_target(line, procedure, vault):
    # Ignore the if condition. This is not an ideal solution, but the easier for the moment
    line = BasicFunctions.remove_if_condition(line)
    if line.count("::"):
        line = line.split("::")[1]
    pointer, target = line.strip().split("=>")
    pointer = pointer.strip()
    target = target.strip()

    # Sometimes indexing uses SIZE intrinsic than makes it impossible to find the variable pt4d(1:SIZE(ptab, dim=1))
    if NatureDeterminer.has_intrinsics(pointer):
        pointer = NatureDeterminer.strip_from_intrinsic(pointer)[0]
    ptr = get_variable(pointer, procedure, vault)

    if target.lower().replace(" ", "") == "null()":
        return [None, None]
    else:
        # Is a pointer to procedure
        if ptr.type == 'procedure':
            try:
                from AutoRPE.UtilsRPE.BasicStructures import Function
                target = vault.get_procedure_by_name(target.strip(), procedure.module.name)
                # It is a pointer to function, i.e. a variable is associated with it
                return ptr, target
            except Error.ProcedureNotFound:
                # It is a pointer to subroutine
                return [None, None]
        else:
            target = get_variable(target, procedure, vault)
    return ptr, target


# Retrieve variable corresponding to a structure member
def get_structure_member(contents, procedure, vault):
    root = contents.split("%")[0]

    # Find root variable
    root_variable = vault.get_variable_by_name(root, procedure)

    # Search for root derived type definition
    root_derived_type = [dt for dt in vault.derived_types if dt.name.lower() == root_variable.type.lower()][0]
    if root_derived_type.is_external:
        return None
    layers = contents.split("%")[1:]
    for l in layers:
        l = remove_indexing(l)
        try:
            layer = [v for v in root_derived_type.variables if v.name.lower() == l.lower()][0]
        except IndexError:
            # It means that mother type is an extended type
            try:
                layer = [v for v in root_derived_type.extension.variables if v.name.lower() == l.lower()][0]
            except IndexError:
                raise Error.MemberNotFound
            if layer:
                pass
            else:
                raise Error.VariableNotFound("Variable %s not found", contents)

        try:
            root_derived_type = [dt for dt in vault.derived_types if dt.name.lower() == layer.type.lower()][0]
        except IndexError:
            if l == contents.split("%")[-1]:
                pass
            else:
                raise Error.VariableNotFound("Variable %s not found", contents)

    return layer


def get_type_of_contents(contents, procedure, vault):
    contents = contents.replace(" ", "").strip()

    # Basic check
    if contents[0] == "-" or contents[0] == "+":
        contents = contents[1:]

    if contents.lower()[0] in ["'", '"'] or contents.lower()[:4].strip() == 'trim':
        return "char"
    if contents.count("'") or contents.count("\""):
        return "char"

    if NatureDeterminer.is_logical_comparison(contents):
        return "logical"

    # Check if is an hardcoded array, and return the content type
    if NatureDeterminer.is_an_hardcoded_array(contents):
        contents = Splitter.split_hardcoded_array(contents)[0]
        return get_type_of_contents(contents, procedure, vault)

    # After checking it is not an array we remove parenthesis
    if BasicFunctions.is_all_inside_parenthesis(contents):
        contents = BasicFunctions.remove_outer_brackets(contents)

    # Check if is a call to function including intrinsic
    if CallManager.is_call_to_function(contents, procedure, vault):
        function_type = get_function_type(contents, procedure, vault)
        if function_type:
            if function_type == "external":
                return None
            else:
                return function_type

    if NatureDeterminer.has_operations(contents):
        split_elements = Splitter.split_elements_in_operations(contents, procedure=procedure, vault=vault)
        types = set(get_type_of_contents(element, procedure, vault) for element in split_elements)
        # Remove any possible None
        types = set([t for t in types if t not in [None, 'None']])
        if len(types) > 1:
            # Different types were used in the operation
            return BasicFunctions.combine_types(set(types))
        else:
            # Operands were all of the same type
            return types.pop()

    if is_keyword_argument(contents):
        # Removing before first =
        contents = contents[contents.find("=") + 1:].strip()
        try:
            return get_type_of_contents(contents, procedure, vault)
        except Error.VariableNotFound:
            warnings.warn('External variable %s used' % contents)
            return None

    # Checks if it is a number
    num_type = NatureDeterminer.is_a_number(contents)
    if num_type:
        return num_type

    # Remove the indexing from array and DT:
    #   ght_abl <-- ght_abl(2:jpka)
    #   todelay%z1d <-- todelay(ji)%z1d(:)
    contents = remove_indexing(contents)

    # Search if is a simple variable
    if re.search(RegexPattern.variable, contents):
        var = vault.get_variable_by_name(contents, procedure)
        return var.type

    # Check if is a derived data type
    if contents.count("%"):
        if contents.split("%")[-1] == "val":
            return VariablePrecision.real_id['dp']
        member = get_structure_member(contents, procedure, vault)
        if member:
            return member.type
        else:
            # Case of an external data type
            return None

    raise Error.TypeOfContent("Type of content for the argument %s: not found", contents)


def get_intrinsic_type(function_call, procedure, vault):
    import AutoRPE.UtilsRPE.SourceManager as SourceManager
    function_name = CallManager.find_called_function(function_call)
    try:
        intrinsic_type = SourceManager.list_of_intrinsics[function_name.lower()]
    except KeyError:
        return None
    if intrinsic_type == 'content_type':
        if function_name.lower() == 'real':
            return get_type_of_real_cast(function_call)
        return get_type_of_contents(CallManager.find_call_arguments(function_call)[0], procedure, vault)
    elif intrinsic_type == 'wp':
        return VariablePrecision.real_id['wp']
    elif intrinsic_type in [None, 'None']:
        raise ValueError("Please add the type specification for this intrinsic: %s" % function_name)
    else:
        return intrinsic_type


def get_function_type(function_call, procedure, vault):
    # Check if is an intrinsic
    if NatureDeterminer.has_intrinsics(function_call):
        return get_intrinsic_type(function_call, procedure, vault)
    else:
        # Must be a function stored in vault
        function_name = CallManager.find_called_function(function_call)
        try:
            function = vault.get_procedure_by_name(function_name)
        except Error.ProcedureNotFound:
            return None
        if isinstance(function, BasicStructures.Interface):
            called_arguments = CallManager.find_call_arguments(function_call)
            function = get_interface_from_arguments(called_arguments, function, procedure, vault)
        return function.type


def get_dimension_of_contents(contents, procedure=None, vault=None):
    # TODO: Improve this function to work properly in all cases.
    contents = BasicFunctions.remove_literals(contents)

    # Its a Char
    if contents.count("'") or contents.count('"') or contents.lower().count("trim"):
        return 0
    if contents.lower().count(".false.") or contents.lower().count(".true."):
        return 0
    if is_keyword_argument(contents):
        return get_dimension_of_contents(contents.split("=")[1], procedure, vault)
    # Matches arrays in the form of (/a,b,c/) or [a,b,c]
    if NatureDeterminer.is_an_hardcoded_array(contents):
        return 1
    # Adding a workaround for casts
    if CallManager.is_call_to_function(contents, procedure, vault):
        fun = CallManager.find_called_function(contents)
        if fun.lower() in ["real", "rpe", "sqrt", "min", "max"]:
            return get_dimension_of_contents(CallManager.find_call_arguments(contents)[0], procedure, vault)
        elif fun.lower() == "sum" or fun.lower() == "maxval":
            if contents.lower().count("dim="):
                dim = get_dimension_of_contents(CallManager.find_call_arguments(contents)[0], procedure, vault)
                if dim:
                    return dim - 1
                else:
                    return dim
            else:
                contents = CallManager.find_call_arguments(contents)[0]
        elif fun.lower() == "count":
            return 0
        else:
            # it was not a cast?
            pass

    if NatureDeterminer.has_operations(contents):
        elements = Splitter.split_elements_in_operations(contents)
        dimensions = [get_dimension_of_contents(e, procedure=procedure, vault=vault) for e in elements]
        dimensions = [d for d in dimensions if d is not None]
        if not dimensions:
            dimensions = [0]
        return max(dimensions)
    if NatureDeterminer.is_a_number(contents):
        return 0

    # Count how many slice are considered
    if contents.count(":"):
        return contents.count(":")

    try:
        var = get_variable(contents, procedure=procedure, vault=vault)
        # Checking if we have the var + slice specification, like sdjf%fdta(2,2,1,2) or psgn(jf)
        # then it must be an scalar
        if re.search(r"%s *\(.*\)" % var.name, contents):
            return 0
        return var.dimension
    except AssertionError:
        return None


def get_variable_id(vault, variable_name, procedure_name=None, module_name=None, ):
    # Small function created to easily find the id of variables that appear in the code
    vs = [v for v in vault.variables if v.name == variable_name]

    if module_name is not None:
        vs = [v for v in vs if v.procedure.module.name == module_name]
    if procedure_name is not None:
        vs = [v for v in vs if v.procedure.name == procedure_name]

    if len(vs) == 1:
        print("Variable found! ID: %i" % vs[0].id)
        return vs[0].id
    elif len(vs) > 1:
        print("More than one variable!")
        return None
    else:
        print("Not a single variable found!")
        return None


def get_interface_from_arguments(arguments, interface, procedure=None, vault=None):
    argument_types = [get_type_of_contents(argument, procedure, vault) for argument in arguments]
    argument_dim = [get_dimension_of_contents(argument, procedure, vault) for argument in arguments]
    keyword_arguments = [_x.split("=")[0].strip() if is_keyword_argument(_x) else False for _x in arguments]
    try:
        get_interface_from_key(interface, argument_types, argument_dim)
    except Error.InterfaceNotFound:
        pass
    return get_interface_from_arguments_and_info(argument_types, interface, argument_dim, keyword_arguments)


def get_interface_from_key(interface, argument_types, argument_dimensions):
    key = BasicStructures.generate_dict_key(argument_types, argument_dimensions, uniform=True)
    if key in interface.agnostic_key:
        return interface.agnostic_key[key]
    raise Error.InterfaceNotFound("No matching routine found for interface : %s" % interface.name)


def get_interface_from_arguments_and_info(argument_types, interface, argument_dimensions, keyword_arguments):
    # This function is intended to take the argument info (types and dimensions) and return the subroutine
    # inside the interface that matches it.
    number_of_arguments = len(argument_types)

    for procedure in interface.subprogram:

        mandatory_arguments = [argument for argument in procedure.dummy_arguments if not argument.is_optional]

        if len(procedure.dummy_arguments) < number_of_arguments:
            continue
        if len(mandatory_arguments) > number_of_arguments:
            continue
        match = True
        for index, argument_type in enumerate(argument_types):
            if match is False:
                break
            keyword = keyword_arguments[index]
            if keyword:
                var = procedure.get_variable_by_name(keyword)
                if var:
                    dummy_type = var.type
                    dummy_dimension = var.dimension
                else:
                    dummy_type = None
                    dummy_dimension = None
                    match = False
            else:
                dummy_type = procedure.get_variable_by_position(index).type
                dummy_dimension = procedure.get_variable_by_position(index).dimension
            if not VariablePrecision.type_are_equivalent(dummy_type, argument_type):
                if (dummy_type in VariablePrecision.real_id and argument_type == "rpe_var") or (
                        dummy_type == "rpe_var" and argument_type in VariablePrecision.real_id):
                    pass
                else:
                    match = False
            if argument_dimensions is not None:
                # At this point we'll just ignore char variables.
                if argument_type == "char":
                    continue
                argument_dimension = argument_dimensions[index]
                if argument_dimension is not None and argument_dimension != dummy_dimension:
                    match = False

        if match:
            return procedure
    raise Error.InterfaceNotFound("No matching routine found for interface : %s" % interface.name)


def is_keyword_argument(argument):
    pattern = "^[^()'\"=]*=[^=]"
    if re.search(pattern, argument):
        return True
    else:
        return False


def get_type_of_real_cast(string):
    # Leave only the arguments
    arguments = CallManager.find_call_arguments(string)
    if len(arguments) == 1:
        return VariablePrecision.real_id['wp']
    elif len(arguments) == 2:
        argument, type = arguments
        return VariablePrecision.lookup_table[type]


def remove_indexing(content):
    # If the variable is already a simple name
    if re.match(RegexPattern.variable, content):
        return content
    # Avoid problem with structure used as index
    content = content.replace("%", "XXX")
    # Used to avoid recursion error
    clean_content = content
    # Replace hardcoded array
    hard_code_array = re.search(RegexPattern.hardcoded_array, clean_content)
    if hard_code_array:
        clean_content = clean_content.replace(hard_code_array.group(), "")
        # Remove possible container of hardcoded array
        clean_content = clean_content.replace("()", "")
    match_array_brackets = re.findall(RegexPattern.match_array_brackets, clean_content)
    for brakets in match_array_brackets:
        clean_content = clean_content.replace(brakets, "")
    # Recursively remove brackets
    if not re.search(RegexPattern.variable, clean_content) and not re.search(RegexPattern.structure, clean_content):
        # This check avoid infinite recursion, if the function has not changed anything returns
        if clean_content != content:
            return remove_indexing(clean_content)

    # Reinsert structure markers
    clean_content = clean_content.replace("XXX", "%")
    return clean_content
