import AutoRPE.UtilsRPE.SourceManager as SourceManager
import AutoRPE.UtilsRPE.BasicFunctions as BasicFunctions
import AutoRPE.UtilsRPE.RegexPattern as RegexPattern
import AutoRPE.UtilsRPE.VariablePrecision as VariablePrecision
import AutoRPE.UtilsRPE.Error as Error

import re


def is_function_declaration_statement(line):
    # Remove matches inside write(*,*) statements
    line = BasicFunctions.remove_literals(line)
    # Check for function declaration syntax
    is_start = re.search(RegexPattern.function_declaration, line, re.MULTILINE | re.IGNORECASE)
    if is_start:
        return True
    return False


def is_end_function_statement(line):
    is_end = re.match(RegexPattern.end_function_declaration, line, re.I)
    if is_end:
        return True
    return False


# Check if a string is a number and returns the type
def is_a_number(string):

    # Check if is just white spaces
    if not string.strip():
        return False
    else:
        string = string.lower().strip()

    # Remove unary operator if present
    if string[0] == '-' or string[0] == '+':
        string = string[1:]

    if has_intrinsics(string):
        return False

    # Checks for number in exponential notation
    exp = re.match("^" + RegexPattern.number_exp_DPE_notation + "$", string)
    # Checks for number with precision specified
    real_with_prec = re.match(RegexPattern.real_number_with_precision, string)
    if exp or real_with_prec:
        # In case the number is a hardcoded real with a type specification,
        # check the lookup table to know to which type it corresponds and return it
        try:
            new_type = VariablePrecision.lookup_table[string.split("_")[1]]
        except IndexError:
            return VariablePrecision.real_id['wp']
        return new_type
    integer_with_prec = re.match(RegexPattern.integer_number_with_precision, string)
    if integer_with_prec:
        return "integer"
    # Check for pure integer and real numbers
    try:
        float(string)
        if string.isdigit():
            return "integer"
        else:
            return VariablePrecision.real_id['wp']
    except ValueError:
        return False


def is_an_hardcoded_array(string):
    match = re.match(RegexPattern.hardcoded_array, string)
    if match:
        return True
    return False


def is_pointer_assignment(string):
    pattern = ".* => .*"
    mask = BasicFunctions.literal_mask(string)
    match = re.search(pattern, string, flags=re.I)
    if match and not mask[string.index("=>")]:
        return True
    return False


def is_logical_comparison(string):
    for element in comparison_operators:
        if string.lower().count(element):
            return True
    return False


def has_operations(string):
    import AutoRPE.UtilsRPE.CallManager as CallManager
    operations = ["+", "-", "*", "/"]
    # Exclude key arguments
    # TODO: Deal with "list comprehension" i.e. : ' (/ (zshift + REAL(ji,wp), ji = 1, jpi_crs*jpj_crs) /)'
    # TODO: This approach does not work
    keyword_pattern = "^ *\w+ *=.*"
    if re.match(keyword_pattern, string, re.I):
        return False
    # Check if string is just white spaces
    if not string.strip():
        return False
    else:
        string = string.lower().replace(" ", "")

    if string.count("'") or string.count('"'):
        return False
    if CallManager.is_call_to_function(string):
        return False
    # Remove the unary operators
    if string[0] == '-' or string[0] == '+':
        string = string[1:]

    # Replace char that can cause false positive
    string = string.replace("=-", 'EQUAL_MIN')
    string = string.replace("(/", "BEGIN_VEC(")
    string = string.replace("/)", ")")

    # Substitutes arrays and structures, then deletes useless brackets
    string = remove_arrays_and_structures(string)

    # Check for exponential, regex is needed to be sure is not just one of the variables that ends with 'e'
    exp_DPE = re.search(RegexPattern.number_exp_DPE_notation, string)
    if exp_DPE:
        string = string.replace(exp_DPE.group(), 'EXPONENTIAL_DPE')

    for op in operations:
        if op in string:
            return True
    return False


def has_intrinsics(string):
    for key in list_of_key:
        if string.lower().count(key):
            key_match = r"\s*[ ,(+-]%s *= *\w" % key
            string = re.sub(key_match, "", string, re.I)

    # Split parenthesis opening
    if not string.count("("):
        return False

    possible_intrinsics = re.findall(r'\s*\b\w+\b\s*\(', string, re.I)

    for name in possible_intrinsics:
        # Remove last parenthesis and leading spaces
        name = name[:-1].strip()
        if name.lower() in SourceManager.list_of_intrinsics:
            return name
    return None


def strip_from_intrinsic(string):
    import AutoRPE.UtilsRPE.CallManager as CallManager
    intrinsic = has_intrinsics(string)

    while intrinsic:
        # Get the intrinsic call
        intrinsic = BasicFunctions.close_brackets(string[re.search(r'\b%s\b\(' % intrinsic, string, re.I).start():])
        called_arg = CallManager.find_call_arguments(intrinsic)
        arguments_to_remove = []
        for arg in called_arg:
            if arg.split("=")[0] in list_of_key:
                arguments_to_remove.append(arg)
        # Removed key specification
        for arg in arguments_to_remove:
            called_arg.remove(arg)
        # Rejoin pure arguments
        if len(called_arg) > 1:
            string = string.replace(intrinsic, "("+",".join(called_arg)+")", re.I)
        else:
            string = string.replace(intrinsic, called_arg[0], re.I)
        if BasicFunctions.is_all_inside_parenthesis(string):

            string = BasicFunctions.remove_outer_brackets(string)
        # Check for further intrinsics
        intrinsic = has_intrinsics(string)
    # Check if is a cast with precision and the precision was not specified via key
    called_arg = CallManager.find_call_arguments("(" + string + ")")
    if len(called_arg) == 2:
        if called_arg[1] in VariablePrecision.lookup_table:
            return [called_arg[0]]
        # It is the dimension ?
        if is_a_number(called_arg[1]):
            return [called_arg[0]]
    if len(called_arg) > 1:
        raise Error.VariableNotFound("The string %s is not a variable" % string)
    return called_arg


# PRIVATE

# Substitutes arrays and structures, then deletes useless brackets
# Used in has_operations
def remove_arrays_and_structures(string):
    replacement_index = 0
    arrays = True
    replacement_list = []
    # Checks for all array pattern, and substitute with string
    while arrays:
        arrays = re.findall(RegexPattern.array, string)
        for array in arrays:
            string = string.replace(array, "REPLACEMENT" + str(replacement_index), 1)
            replacement_list.append(array)
            replacement_index += 1

        # Substitute structures with string
        structures = re.findall(r"\w+%\w+", string)
        for struct in structures:
            string = string.replace(struct, "REPLACEMENT" + str(replacement_index))
            replacement_list.append(struct)
            replacement_index += 1
            arrays = True
    return string


list_of_key = ["dim", "mask", "kind", "back"]
comparison_operators = [">=", "<=", "==", "/=", ">", "<", ".not.", ".or.", ".and.", ".false.", ".true."]
