import AutoRPE.UtilsRPE.BasicStructures as BasicStructures
import AutoRPE.UtilsRPE.BasicFunctions as BasicFunctions
import AutoRPE.UtilsRPE.CallManager as CallManager
import AutoRPE.UtilsRPE.NatureDeterminer as NatureDeterminer
import AutoRPE.UtilsRPE.AttributeParser as AttributeParser
import AutoRPE.UtilsRPE.VariablePrecision as VariablePrecision
import AutoRPE.UtilsRPE.RegexPattern as RegexPattern
import AutoRPE.UtilsRPE.Error as Error
import re


def parse_sources(path_to_model_sources, vault):
    files = BasicFunctions.load_sources(path_to_model_sources)
    # Parsing subprograms and storing info in Module structures.
    n_of_files = len(files)
    print("\n")
    for index, file in enumerate(files):
        print("\rParsing subprograms and variables %i/%i " % (index + 1, n_of_files), end="")
        module = BasicStructures.Module(file.path)
        module.main = BasicStructures.Procedure("main", module)
        find_subroutines(module, vault)
        find_functions(module, vault)
        find_module_variables(module)
        vault.modules.append(module)
        # Add it after the main variables have been found, so as to add them to the variable dictionary as well
        vault.add_procedure(module.main)

    # Adding info on internal declaration of subroutines
    for procedure in vault.procedures_dictionary:
        declares = vault.procedures_dictionary[procedure][0].declares
        if declares:
            for declared_sbr in declares:
                vault.procedures_dictionary[declared_sbr][0].scope = vault.procedures_dictionary[procedure][0]

    print("\nParsing Interfaces\n")
    # Parsing the interfaces and storing in Module.
    for index, module in enumerate(vault.modules):
        find_interfaces(module, vault)

    # # Fill vault variable list:
    for procedures in vault.procedures_dictionary.values():
        for proc in procedures:
            if not isinstance(proc, BasicStructures.Interface):
                for var in proc.variables:
                    vault.variables.append(var)

    # Fill blocks, and store used modules info
    for index, module in enumerate(vault.modules):
        module.update_blocks(vault)
        # Find used module and store
        find_used_modules(module, vault)

    # After having all the used parsed, propagate the inheritance, and add public var to available vars
    print("Expanding use statements\n")
    vault.expand_use_statement()

    print("Parsing derived types\n")
    # Parsing the derived types
    for module in vault.modules:
        find_types(module, vault)

    # Store info of derived types in vault procedures
    for dt in vault.derived_types:
        for var in dt.variables:
            vault.variables.append(var)

    # Now that the dictionary if full, fill the blocks info
    for module in vault.modules:
        module.update_blocks(vault)

    # Add to variables info about their belongings in namelist, to be used when inserting the %sbits info
    for module in vault.modules:
        find_namelist_variable(module, vault)

    # Now that vault.derived_types is full add extension to derived types if any
    vault.expand_derived_types()

    # Store the external used data types
    vault.store_external_data_structures()

    # Stores the external function before subroutines, because this can be used as arguments of the second
    for module in vault.modules:
        find_external_functions(module, vault)

    # Stores the external call found in the code
    for module in vault.modules:
        find_external_subroutines(module, vault)

    # Passing to pointed objects the actual function/subroutine they points to
    process_pointers(vault)
    n_func = len([p for p in vault.procedures_dictionary.values() if isinstance(p[0], BasicStructures.Function)])
    n_sbr = len([p for p in vault.procedures_dictionary.values() if isinstance(p[0], BasicStructures.SubRoutine)])
    n_interfaces = len([p for p in vault.procedures_dictionary.values() if isinstance(p[0], BasicStructures.Interface)])
    print("number of functions " + str(n_func) + "\n" +
          "number or subroutines " + str(n_sbr) + "\n" +
          "number or interfaces " + str(n_interfaces) + "\n" +
          "number or main " + str(len(vault.modules)) + "\n" +
          "number or procedures " + str(len(vault.procedures_dictionary)) + "\n" +
          "number of variables " + str(len(vault.variables)) + "\n" +
          "number or derived_types " + str(len(vault.derived_types)) + "\n")

    return vault


def get_dummy_arguments(module, index):
    # Definition: Given a line number, returns the subroutine name and the dummy arguments (in case they exist)
    # Input: integer with line number
    # Output: tuple with routine name and list of dummy arguments
    line = module.lines[index]
    # Workaround to solve the problem that appears when there's a function declaration with (kind=...)
    if line.lower().count("function "):
        if line.find("(") < line.lower().find("function "):
            line = line[line.lower().find("function "):]

    # If the subroutine declaration does not have any parenthesis, this routine does not have any argument
    if not line.count("("):
        name = line.split()[1].strip()
        dummy_arguments = []
    # If otherwise it has dummy arguments, find them
    elif line.lower().count("result") or line.lower().count("bind("):
        name = line[:line.find("(")].split()[-1]
        dummy_arguments = []
        dummy_arguments += CallManager.strip_call_arguments(BasicFunctions.close_brackets(line))
        # I don't know if we should record the result variable with the other variables
        # result_start = line.lower().find("result")
        # dummy_arguments += strip_call_arguments(line[result_start:])
    else:
        name = line[:line.find("(")].split()[-1]
        dummy_arguments = CallManager.strip_call_arguments(line)
    dummy_arguments = [x for x in dummy_arguments if x.strip()]
    return name, dummy_arguments


def find_subprogram_declaration(module, _subprogram_start, _subprogram_end, obj_to_find=RegexPattern.variable_declaration):
    # Definition: Given the line numbers where a routine starts and ends,
    # return list of strings containing declarations
    # TODO: manage cases in which a routine contains other functions declarations.

    stop_index = [index
                  for index, line in enumerate(module.lines[_subprogram_start:_subprogram_end + 1])
                  if re.search(RegexPattern.contains_declaration, line, re.I) or re.search(RegexPattern.interface_declaration, line, re.I)]
    # A declaration found stop the search there
    if stop_index:
        _subprogram_end = _subprogram_start - 1 + stop_index[0]

    # Lines inside the routine
    _subprogram_lines = module.lines[_subprogram_start:_subprogram_end + 1]

    # Lines that are variable declarations
    _all_declaration = [_l.strip() for _l in _subprogram_lines if
                        re.search(obj_to_find, _l, re.I) and not
                        (re.search(RegexPattern.general_allocation, _l, re.I) or
                         re.search(RegexPattern.string_array_initialization, _l, re.I))]

    return _all_declaration


def add_struct_members(file, structure):
    # Definition: Given the routine name, the name of the dummy arguments and the lines containing declarations,
    # this method finds the attributes of each dummy arguments and adds all the information
    # into the object variable "routines".
    # Input:
    #       -routine_name: string with the routine name
    #       -dummy_arguments: list of strings with argument names
    #       -declarations: list of strings with variable declarations
    try:
        for obj in file.objects:
            if obj.name == "main":
                main = obj
                break
        for member in structure.variables:
            main.add_variable(member)

    except UnboundLocalError:
        print("Main not found for structure %s", structure.name)


def create_subprogram(module, header, routine_name, dummy_arguments, var_declarations, used_modules,
                      subprogram_type):
    if subprogram_type == BasicStructures.Function:
        # Retrieve the ret_val name
        ret_type, ret_val = find_function_type_from_header(header, var_declarations)
        if not ret_val:
            ret_val = routine_name
    elif subprogram_type == BasicStructures.SubRoutine:
        ret_val = None
    else:
        raise ValueError("Header does not correspond to function or routine")

    subprogram = subprogram_type(routine_name, module)

    if len(used_modules):
        subprogram.use_only_variables = [m.lower().split("use")[1].strip() for m in used_modules]
    # Reverse approach
    dummy_arguments = [x for x in dummy_arguments]
    # var_names = [AttributeParser.get_variable_name(var.split('::')[1])[0] for var in var_declarations]
    # for da in dummy_arguments:
    #     if da not in var_names:
    #         print("This variable has a different spelling btw declaration and dummy")
    for declaration in var_declarations:
        pattern = "(.*)::(.*)"
        m = re.search(pattern, declaration)
        # In this case is not a variable declaration
        if declaration.count("contiguous"):
            continue
        _attributes, variable = m.groups()
        # Check if the dummy argument is inside the declaration arguments
        # If it is, take the type, dimension, intent and optional
        _var_type, _dimension, _intent, _optional, \
        _allocatable, is_pointer, is_parameter = AttributeParser.parse_attributes(_attributes, variable)

        variable_name = AttributeParser.get_variable_name(variable)

        if len(variable_name) != 1:
            raise AssertionError("Problem identifying variable declaration.")
        variable_name = variable_name[0]
        _v = BasicStructures.Variable(subprogram, variable_name, _var_type, _dimension, module)
        _v.allocatable = _allocatable
        _v.is_parameter = is_parameter
        # If the variable name is the same of the function or is return statement
        # if obj_type == BasicStructures.Function and _v.name.lower() == ret_val.lower():
        if ret_val is not None and _v.name.lower() == ret_val.lower():
            _v.is_retval = True
            subprogram.type = _var_type
        _v.is_pointer = is_pointer
        if variable_name in dummy_arguments:
            _v.intent = _intent
            _v.is_optional = _optional
            _v.is_dummy_argument = True
            subprogram.dummy_arguments.append(_v)
        subprogram.add_variable(_v)

    # Save the position of the argument
    found_vars = [variable.name for variable in subprogram.variables]

    for index, dummy_argument in enumerate(dummy_arguments):
        if dummy_argument in found_vars:
            subprogram.get_variable_by_name(dummy_argument).position = index
            continue
            # If it's not found its an implicitly declared variable
        if dummy_argument[0] in "ijklmn":
            var_type = "integer"
        else:
            var_type = VariablePrecision.real_id['wp']
        dimension = 0
        intent = None
        optional = None
        _v = BasicStructures.Variable(subprogram, dummy_argument, var_type, dimension, module)
        _v.intent = intent
        _v.is_optional = optional
        _v.is_dummy_argument = True
        _v.position = index
        _v.is_parameter = False
        subprogram.add_variable(_v)
        subprogram.dummy_arguments.append(_v)
    # Order dummy_arguments by position
    subprogram.dummy_arguments.sort(key=lambda x: x.position)
    if subprogram_type == BasicStructures.Function:
        # In case of function  declaration like INTEGER FUNCTION name() without return variable declaration
        if not [var for var in subprogram.variables if var.is_retval]:
            # Add the return value to functions variables
            _v = BasicStructures.Variable(subprogram, ret_val, ret_type, 0, module)
            _v.is_retval = True
            subprogram.add_variable(_v)
            # Specify the type of the function
            subprogram.type = ret_type
    # Save object
    return subprogram


def find_subroutines(module, vault):
    #  Looks into the sourcefile for routine declarations, storing
    #  name, type, dimensions and optionality of the arguments.

    for _index, line in enumerate(module.lines):
        # Look if routine starts with (recursive) subroutine
        sbr_name = re.search(RegexPattern.subroutine_declaration, line, re.IGNORECASE)
        if sbr_name:
            sbr_name = sbr_name.group(1)
            subroutine_start = _index
            # Look for end subroutine statements
            subroutine_end = [index for index, l in enumerate(module.lines)
                              if re.search(r'end\s+subroutine\s+%s\b' % sbr_name, l, re.I)][0]

            # Find name and dummy arguments
            subroutine_name, dummy_arguments = get_dummy_arguments(module, _index)
            # Find declarations
            var_declarations = find_subprogram_declaration(module, subroutine_start, subroutine_end,
                                                           obj_to_find=RegexPattern.variable_declaration)
            used_modules = find_subprogram_declaration(module, subroutine_start, subroutine_end,
                                                       obj_to_find=RegexPattern.use_statement)
            header = module.lines[subroutine_start]
            # Create subroutine objects, with variables and their attribute. Then add it to the module
            subroutine = create_subprogram(module, header, subroutine_name, dummy_arguments, var_declarations,
                                           used_modules, subprogram_type=BasicStructures.SubRoutine)
            # Check if subroutines are defined inside this scope
            subroutine_lines = module.lines[subroutine_start+1:subroutine_end-1]
            internal_sbr = [re.search(RegexPattern.subroutine_declaration, l, re.I)  for l in subroutine_lines
                            if re.search(RegexPattern.subroutine_declaration, l, re.I)]
            if internal_sbr:
                sbr_names = [i_s.group(1) for i_s in internal_sbr]
                subroutine.declares = sbr_names

            vault.add_procedure(subroutine)


def find_external_subroutines(module, vault):
    # Store all the external called subroutines
    for index, line in enumerate(module.lines):
        # Check if there is a call to subroutine in line
        line = BasicFunctions.remove_if_condition(line)
        sbr_call = CallManager.call_to_subroutine(line)

        if sbr_call:
            # Returns a list, but if the code is preprocessed the list will always have length 1
            try:
                CallManager.get_subroutine(line, module.blocks[index], vault)

            except Error.ProcedureNotFound as unknown_subroutines:
                for sbr_name in unknown_subroutines.args[0]:
                    try:
                        vault.procedures_dictionary[sbr_name]
                    except KeyError:
                        vault.procedures_dictionary[sbr_name] = [BasicStructures.SubRoutine(sbr_name, module=None,
                                                                                            is_external=True)]


def find_external_functions(module, vault):
    # Store all the external called subroutines
    for index, line in enumerate(module.lines):
        # Check if there is a call to function in line
        # the replace spaces is a workaround for the regex to work: it contains a lookahead
        line = BasicFunctions.remove_if_condition(line)
        func_call = re.search(RegexPattern.call_to_function, re.sub('\\s\\s+', r' ', line.strip()), re.I)

        if func_call and not (line.count("::") or NatureDeterminer.is_function_declaration_statement(line)):
            # Check if the function called is external
            try:
                CallManager.get_function(line, module.blocks[index], vault)

            except Error.ProcedureNotFound as unknown_functions:
                for func_name in unknown_functions.args[0]:
                    try:
                        vault.procedures_dictionary[func_name]
                    except KeyError:
                        vault.procedures_dictionary[func_name] = [BasicStructures.Function(func_name, module=None,
                                                                                           is_external=True)]


def find_types(module, vault):
    # Looks into the module lines for type declaration
    import AutoRPE.UtilsRPE.RegexPattern as RegexPatter
    # Look into each line
    in_type_declaration = False
    derived_types = []
    type_name = None
    variable_counter = None
    for _index, line in enumerate(module.lines):
        # Look if routine starts with subroutine
        if not in_type_declaration:
            # check if a type has been declared
            m = re.search(RegexPatter.simple_type_declaration, line, re.IGNORECASE)
            if m:
                in_type_declaration = True
                variable_counter = 0
                type_name = m.groups()[0]
                derived_type = BasicStructures.DerivedType(type_name, module)
                derived_types.append(derived_type)

            # Case in which a type is extending another type
            match_extend = re.search(RegexPatter.extended_type_declaration, line, re.IGNORECASE)
            if match_extend:
                in_type_declaration = True
                variable_counter = 0
                type_name = match_extend.groups()[2]
                extended_type = match_extend.groups()[1]
                derived_type = BasicStructures.DerivedType(type_name, module, extended_type)
                derived_types.append(derived_type)
        else:
            # Check if we are exiting a type declaration
            m = re.search(RegexPatter.end_type_statement, line, re.IGNORECASE)
            if m:
                in_type_declaration = False
            else:
                pattern = "(.*)::(.*)"
                m = re.search(pattern, line)
                if m:
                    attributes, variables = m.groups()
                    var_type, dimension, intent, optional, \
                    allocatable, is_pointer, is_parameter = AttributeParser.parse_attributes(attributes, variables)

                    variables = AttributeParser.get_variable_name(variables)
                    for var in variables:
                        v = BasicStructures.Variable(derived_type, var, var_type, dimension, module)
                        v.position = variable_counter
                        v.is_dummy_argument = True
                        v.allocatable = allocatable
                        v.is_parameter = is_parameter
                        v.is_pointer = is_pointer
                        v.is_member_of = type_name
                        derived_type.add_variable(v)
                        derived_type.dummy_arguments.append(v)
                        variable_counter += 1
    # Now that also the variables have been added to data structure we can fill the dictionary
    for dt in derived_types:
        vault.derived_types.append(dt)
        vault.add_procedure(dt)


def find_module_variables(module):
    import AutoRPE.UtilsRPE.RegexPattern as RegexPattern

    line_to_check = True

    pattern = "(.*)::(.*)"
    for line in module.lines:
        if line_to_check:
            # Check if we are entering a type declaration zone
            is_type_declaration = re.search(RegexPattern.general_type_declaration, line, re.IGNORECASE)
            if is_type_declaration or NatureDeterminer.is_function_declaration_statement(line):
                line_to_check = False
                continue
        else:
            # Check if we are exiting a type declaration zone
            is_end_type_declaration = re.search(RegexPattern.end_type_statement, line, re.IGNORECASE)
            if is_end_type_declaration or NatureDeterminer.is_end_function_statement(line):
                line_to_check = True
            # Even if we are inside or exiting a type declaration zone, we have to jump to next line
            continue

        if line.lower().count("contains"):
            break
        if line.lower().count("end program"):
            break
        match = re.search(pattern, line)
        if match:
            attributes, variables = match.groups()
            var_type, dimension, intent, optional, \
            allocatable, is_pointer, is_parameter = AttributeParser.parse_attributes(attributes, variables)
            variables = AttributeParser.get_variable_name(variables)
            if var_type:
                for variable in variables:
                    v = BasicStructures.Variable(module.main, variable, var_type, dimension, module)
                    v.allocatable = allocatable
                    v.is_pointer = is_pointer
                    v.is_parameter = is_parameter
                    module.main.add_variable(v)
                    module.variables.append(v)


def find_namelist_variable(module, vault):
    if module.text.lower().count('namelist'):
        for index, line in enumerate(module.lines):
            if line.lower().strip().find("namelist") == 0:
                # Retrieve all variables in the namelist and set the module and name of the namelist
                namelist_variables = [var.strip() for var in line.split("/")[2].split(",")]
                namelist_name = re.search(r'namelist\s*/(.*)/', line, re.I).group(1)
                for nv in namelist_variables:
                    var = vault.get_variable_by_name(nv, module.blocks[index])
                    var.appears_in_namelist = [module.module_name, namelist_name]


def find_functions(module, vault):
    # Definition: Looks into the sourcefile for routine declarations, storing the name, type,
    # dimensions and optionality of the arguments.
    for _index, line in enumerate(module.lines):
        if NatureDeterminer.is_function_declaration_statement(line):
            function_start = _index
            # Look for end subroutine statements
            function_end = BasicFunctions.find_next_end(module.lines, _index, "end function")

            # Find name and dummy arguments
            function_name, dummy_arguments = get_dummy_arguments(module, _index)
            # Find declarations
            variables_declarations = find_subprogram_declaration(module, function_start, function_end, obj_to_find="::")
            used_modules = find_subprogram_declaration(module, function_start, function_end, obj_to_find="^use\s")
            # Parse everything and store
            header = module.lines[function_start]
            function = create_subprogram(module, header, function_name, dummy_arguments, variables_declarations,
                                         used_modules, subprogram_type=BasicStructures.Function)
            vault.add_procedure(function)


def find_interfaces(module, vault):
    # Definition: Looks into the sourcefile for interface declarations, storing the name.
    for _index, _line in enumerate(module.lines):
        if _line.strip().lower().find("interface") == 0:
            interface_start = _index
            interface_end = BasicFunctions.find_next_end(module.lines, _index, "end interface")
            try:
                interface_name = _line.split()[1].strip()
            except IndexError:
                # interface to an external procedure or a procedure used in a procedure with the DEFERRED attribute
                continue
            interfaced_procedure = []
            for interface_line in module.lines[interface_start:interface_end + 1]:
                if interface_line.strip().lower().find("module procedure") == 0:
                    line_procedure = interface_line.replace("module procedure", "")
                    line_procedure = line_procedure.replace("MODULE PROCEDURE", "")
                    line_procedure = [_r.strip() for _r in line_procedure.split(",")]
                    interfaced_procedure += line_procedure
                elif interface_line.strip().lower().strip().find("subroutine") == 0:
                    subroutine_name = interface_line[:interface_line.find("(")].split()[1]
                    interfaced_procedure += [subroutine_name]

            interface = BasicStructures.Interface(interface_name, module)
            for ip in interfaced_procedure:
                interface.add_procedure(vault.procedures_dictionary[ip][0])
                interface.add_key(vault.procedures_dictionary[ip][0])
            vault.add_procedure(interface)
            vault.interfaces_dictionary[interface.name] = interface


def find_used_modules(module, vault):
    for index, line in enumerate(module.lines):
        line = line.strip().lower()
        # First line will module name, which we want to skip
        if line.lower().find("module ") == 0:
            continue
        # No need to parse the entire file
        if not line.find("use ") == 0:
            break

        module_name = line.split()[1]

        use_only_procedure = []
        if line.count(","):
            use_only_procedure = line.split(":")[1].lower().replace("only", "")
            use_only_procedure = use_only_procedure.lower().replace(":", "").strip()
            use_only_procedure = use_only_procedure.split(",")
        if module_name.count(","):
            # Remove the only information
            module_name = module_name.split(",")[0].strip()

        # Sometimes use statement are repeated
        # Add to the module the used modules
        used_module = vault[module_name]
        if used_module not in module.uses:
            # If the module is None it means it is an external import
            if used_module is not None:
                module.uses.append(used_module)

        # Look for ?
        pattern = "(\w+) *=> *(\w+)"
        matches = re.findall(pattern, line)
        if matches:
            for match in matches:
                pointer, target = match
                target_var = vault.module.uses[-1].main().get_variable_by_name(target)
                new_var = BasicStructures.Variable(module.blocks[index], pointer, target_var.type,
                                                   target_var.dimension)
                module.blocks[index].add_variable(new_var)


def find_function_type_from_header(header, var_declaration):
    # Get the function name
    fun_name = header.lower().split('function')[1]
    fun_name = fun_name.split("(")[0]
    fun_name = fun_name.replace("(", "")
    fun_name = fun_name.strip()

    # Check if a return val is declared
    if header.lower().count("result"):
        ret_val = header.lower().split('result')[1]
        ret_val = ret_val.replace('results', '')
        ret_val = ret_val.replace('(', '')
        ret_val = ret_val.replace(')', '').strip()
    else:
        ret_val = fun_name

    # Check if the return_val typ is specified among the function variables
    declaration_of_retval = [v.split("::")[1].lower().strip() for v in var_declaration
                             if v.split("::")[1].lower().strip() in [fun_name, ret_val]]
    if declaration_of_retval:
        return None, declaration_of_retval[0]
    # The type is specified in front of the function
    pattern = "(.*)function"
    match = re.search(pattern, header.strip(), re.IGNORECASE)
    fun_type = remove_brackets_content(match.groups(0)[0].strip().lower())
    if fun_type == "real":
        real_declaration = match.group(1)
        real_type_pattern = r"real\((.*)\)"
        nt = re.match(real_type_pattern, real_declaration, re.I)
        try:
            fun_type = VariablePrecision.lookup_table[nt.group(1)]
        except AttributeError:
            fun_type = VariablePrecision.real_id['wp']

    if match.group(0).count('rpe_var'):
        return 'rpe_var', fun_name
    else:
        return fun_type, fun_name


def remove_brackets_content(string):
    op = [i for i, x in enumerate(string) if x == "("]
    cp = [i for i, x in enumerate(string) if x == ")"]
    _new_line = []
    for n in range(len(string)):
        # If comma index n is outside parenthesis (number of openings equal number of closings) substitute
        if len([x for x in op if x <= n]) == len([x for x in cp if x < n]):
            _new_line += [string[n]]
    return "".join(_new_line)


def process_pointers(vault):
    for var in vault.variables:
        if var.is_pointer:
            if var.is_pointer.count("type"):
                pointer_name = var.is_pointer.replace("type", "")
                pointer_name = BasicFunctions.remove_outer_brackets(pointer_name)
                search_array = vault.derived_types
            elif var.is_pointer.count("class"):
                # Same of type, but for compiler reason they have to be declared as classes
                pointer_name = var.is_pointer.replace("class", "")
                pointer_name = BasicFunctions.remove_outer_brackets(pointer_name)
                search_array = vault.derived_types
            elif var.type == "procedure":
                pointer_name = var.is_pointer
                search_array = list(vault.procedures_dictionary)
            else:
                # It is a pointer to real or a pointer to integer
                continue
            pointed_obj = [obj for obj in search_array if obj.name == pointer_name]
            if len(pointed_obj) > 1:
                raise Error.DuplicatedObject
            else:
                var.is_pointer = pointed_obj
            continue
