# This file contains the definition of a Job object.
# Which is a python structure to control the submission and status of a remote job with few other capabilities.


class Job(object):
    def __init__(self, analysis_set, forced_variables=None, reduced_precision_level=10, communicator=None, vault=None,
                 template=None, result_filename=None):
        # Import some stuff
        from AutoRPE.UtilsWorkflow.Utils import generate_jobscript
        # Init variables
        self.vault = vault
        self.analysis_set = analysis_set
        if forced_variables is not None:
            self.forced_variables = forced_variables[:]
        else:
            self.forced_variables = None
        self.communicator = communicator
        self.job_name = "BinarySearch%i" % self.hash
        self.remote_scratch = self.communicator.remote_scratch
        self.remote_logdir = "%s/LOGS" % self.remote_scratch
        self.remote_namelist_dir = "%s/Namelists" % self.remote_logdir
        self.remote_rundir = "%s/Test%i" % (self.remote_scratch, self.hash)

        self.remote_jobscript_path = "%s/test_%i.cmd" % (self.remote_logdir, self.hash)
        self.remote_namelist = "%s/test_namelist_%i" % (self.remote_namelist_dir, self.hash)
        self.local_namelist = "NewWorlds/namelist_precisions"

        self.template = template
        self.result_filename = result_filename
        # parameters to replace:
        self.parameters = {"JOBNAME": self.job_name,
                           "RUNDIR": self.remote_rundir,
                           "LOGDIR": self.remote_logdir,
                           "NAMELIST_PATH": self.remote_namelist,
                           "NUMPROC": "48"
                           }

        self.jobscript = generate_jobscript(self.template, self.parameters)

        self.elapsed_time = None
        self.job_id = None
        self._result = None

        self.reduced_precision_level = reduced_precision_level

        # The default status is PENDING
        self._status = "PENDING"

    @property
    def variable_set(self):
        if self.forced_variables is not None:
            return self.analysis_set + self.forced_variables
        else:
            return self.analysis_set

    def variables(self):
        return [self.vault.get_variable_by_id(_x) for _x in set(self.variable_set)]

    @property
    def remote_status(self):
        from AutoRPE.UtilsWorkflow.Utils import job_status
        _status, _time = job_status(self.job_id, self.communicator)
        _status = _status.strip()
        return _status

    @property
    def status(self):
        return self._status

    @status.setter
    def status(self, new_value):
        possible_status = ["PENDING", "RUNNING", "ASSERTION_PENDING", "SUCCESS", "FAILED", "SUSPENDED"]
        # The list of possible status shall cover all the possible situations in which a job can be found.
        # PENDING: The job has been created and ready to be launched.
        # A PENDING JOB CAN ONLY HAVE SUSPENDED ANCESTORS AND SUCCESS OR FAILED DESCENDANTS
        # RUNNING: The job is RUNNING.
        # A RUNNING JOB CAN ONLY HAVE SUSPENDED ANCESTORS AND SUCCESS OR FAILED DESCENDANTS.
        # SUCCESS: A job that has been completed with successful results and asserted.
        # A SUCCESS job can not have both descendants FAILED
        # FAILED: A job that has been completed with unsuccessful results and that can not be
        # split ( Already was or a single variable set).
        # SUSPENDED: A job that has been completed with unsuccessful results that has been split and depends on the
        # results of descendant jobs.
        # SUSPENDED ancestors must be SUSPENDED.
        # At least one direct descendant should be different from FAILED or COMPLETED
        if new_value in possible_status:
            self._status = new_value
        else:
            raise AssertionError("Unknown status")

    def result(self):
        from AutoRPE.UtilsWorkflow.Utils import obtain_result
        result_filename = self.result_filename
        self._result = obtain_result(self.remote_rundir, self.communicator, result_name=result_filename)
        return self._result

    @property
    def hash(self):
        from AutoRPE.UtilsWorkflow.Utils import hash_from_list
        return hash_from_list(self.variable_set)

    def recover_job_id(self):
        job_id_filename = "%s/job_id.txt" % self.remote_rundir
        self.job_id = self.communicator.sftp.open(job_id_filename).read().decode().strip()

    def check_if_result_already_exists(self):
        self.update_parameters()
        result_path = "%s/%s" % (self.remote_rundir, self.result_filename)
        try:
            self.communicator.sftp.open(result_path)
            # Recovering job_id
            self.recover_job_id()
            return True
        except IOError:
            return False

    def check_if_already_submitted(self):
        self.update_parameters()
        try:
            command = 'squeue --format="%i %j"'
            stdin, stdout, stderr = self.communicator.execute(command)

            output = stdout.read().decode()
            lines = output.split("\n")
            if len(lines) > 1:
                jobs_in_remote_queue = lines[1:]
                for line in jobs_in_remote_queue:
                    if line.strip():
                        job_id, job_name = line.split()
                        if job_name == self.job_name:
                            return job_id

        except IOError:
            return None
        return None

    def submit(self, force=False):
        from AutoRPE.UtilsWorkflow.Utils import generate_namelist, submit_job
        # Check if result already exists
        if force is False:
            completed = self.check_if_result_already_exists()
        else:
            completed = False
        if not completed:
            self.update_parameters()
            # Generate namelist and put it in the remote system
            generate_namelist(self.variable_set, self.local_namelist, self.vault, self.reduced_precision_level)

            self.communicator.put(self.local_namelist, self.remote_namelist)
            # Submit job and obtain job id
            if force is False:
                self.job_id = self.check_if_already_submitted()
            else:
                self.job_id = None
            if self.job_id is None:
                self.job_id = submit_job(self.jobscript, self.remote_jobscript_path, self.communicator)
                print("Job %i successfully submitted to the remote machine with job_id %s" % (self.hash, self.job_id))
            else:
                print("Job %i was already in queue with job_id %s" % (self.hash, self.job_id))
        else:
            pass

    def update_parameters(self):
        from AutoRPE.UtilsWorkflow.Utils import generate_jobscript
        from os.path import isdir
        from os import mkdir
        self.remote_rundir = "%s/Test%i" % (self.remote_scratch, self.hash)
        self.job_name = "BinarySearch%i" % self.hash

        self.remote_jobscript_path = "%s/test_%i.cmd" % (self.remote_logdir, self.hash)
        self.remote_namelist = "%s/test_namelist_%i" % (self.remote_namelist_dir, self.hash)
        namelist_folder = "PrecisionNamelists"
        # TODO: Put it in a better place
        if not isdir(namelist_folder):
            mkdir(namelist_folder)
        self.local_namelist = "%s/namelist_precisions_%i" % (namelist_folder, self.hash)

        # Set of parameters that will be replaced in the job script:
        self.parameters = {"JOBNAME": self.job_name,
                           "RUNDIR": self.remote_rundir,
                           "LOGDIR": self.remote_logdir,
                           "NAMELIST_PATH": self.remote_namelist,
                           "NUMPROC": "48"
                           }
        self.jobscript = generate_jobscript(self.template, self.parameters)

    def compare_with_job(self, other_job, outputs_file_name="", script=""):
        import re
        folder_files_2 = other_job.communicator.list_dir(other_job.remote_rundir)
        filename_2 = [f for f in folder_files_2 if re.search(outputs_file_name, f)][0]
        results_2 = "%s/%s" % (other_job.remote_rundir, filename_2)

        return self.compare_with_file(results_2, outputs_file_name, script)

    def compare_with_file(self, file_path, outputs_file_name="", script=""):
        import re
        folder_files_1 = self.communicator.list_dir(self.remote_rundir)
        filename_1 = [f for f in folder_files_1 if re.search(outputs_file_name, f)][0]
        results_1 = "%s/%s" % (self.remote_rundir, filename_1)

        results_2 = file_path

        with open(script) as input_script:
            remote_script = input_script.read()

        remote_script = remote_script.replace("%FILENAME1%", results_1)
        remote_script = remote_script.replace("%FILENAME2%", results_2)

        local_wrapper_path = "/home/Earth/otintopr/RPE-Canal/AutoRPE/EvaluationScripts/wrapper.sh"
        remote_wrapper_path = "%s/python_wrapper.sh" % self.remote_logdir

        self.communicator.put(local_wrapper_path, remote_wrapper_path)

        remote_script_name = "compare_script.py"
        remote_script_path = "%s/%s" % (self.remote_logdir, remote_script_name)
        with self.communicator.sftp.file(remote_script_path, "w") as remote_script_file:
            remote_script_file.write(remote_script)

        # Change permissions for the wrapper
        command = "chmod 777 %s" % remote_wrapper_path
        stdin, stdout, stderr = self.communicator.execute(command)
        output = str(stdout.read())

        # Execute script through wrapper
        command = "%s %s" % (remote_wrapper_path, remote_script_path)
        stdin, stdout, stderr = self.communicator.execute(command)
        output = str(stdout.read().decode())
        err = str(stderr.read().decode())

        if output.count("not found!"):
            print("File not found")
            return None

        output = [line for line in output.split("\n") if line.strip()]
        results = {}
        for line in output:
            name, value = line.split()
            value = float(value)
            results[name] = value

        if err.strip():
            print(err)
        return results